from nervaluate.evaluator import Evaluator as NewEvaluator
from nervaluate import Evaluator as OldEvaluator
from nervaluate.reporting import summary_report_overall_indices, summary_report_ents_indices, summary_report

def list_to_dict_format(data):
    """
    Convert list format data to dictionary format.
    
    Args:
        data: List of lists containing BIO tags
        
    Returns:
        List of lists containing dictionaries with label, start, and end keys
    """
    result = []
    for doc in data:
        doc_entities = []
        current_entity = None
        
        for i, tag in enumerate(doc):
            if tag.startswith('B-'):
                # If we were tracking an entity, add it to the list
                if current_entity is not None:
                    doc_entities.append(current_entity)
                # Start tracking a new entity
                current_entity = {
                    'label': tag[2:],  # Remove 'B-' prefix
                    'start': i,
                    'end': i
                }
            elif tag.startswith('I-'):
                # Continue tracking the current entity
                if current_entity is not None:
                    current_entity['end'] = i
            else:  # 'O' tag
                # If we were tracking an entity, add it to the list
                if current_entity is not None:
                    doc_entities.append(current_entity)
                    current_entity = None
        
        # Don't forget to add the last entity if there was one
        if current_entity is not None:
            doc_entities.append(current_entity)
        
        result.append(doc_entities)
    
    return result


def generate_synthetic_data(tags, num_samples, min_length=5, max_length=15):
    """
    Generate synthetic NER data with ground truth and predictions.
    
    Args:
        tags (list): List of entity tags to use (e.g., ['PER', 'ORG', 'LOC', 'DATE'])
        num_samples (int): Number of samples to generate
        min_length (int): Minimum sequence length
        max_length (int): Maximum sequence length
    
    Returns:
        tuple: (true_sequences, pred_sequences)
    """
    import random
    
    def generate_sequence(length):
        sequence = ['O'] * length
        # Randomly decide if we'll add an entity
        if random.random() < 0.7:  # 70% chance to add an entity
            # Choose random tag
            tag = random.choice(tags)
            # Choose random start position
            start = random.randint(0, length - 2)
            # Choose random length (1 or 2 tokens)
            entity_length = random.randint(1, 2)
            if start + entity_length <= length:
                sequence[start] = f'B-{tag}'
                for i in range(1, entity_length):
                    sequence[start + i] = f'I-{tag}'
        return sequence
    
    def generate_prediction(true_sequence):
        pred_sequence = true_sequence.copy()
        # Randomly modify some predictions
        for i in range(len(pred_sequence)):
            if random.random() < 0.2:  # 20% chance to modify each token
                if pred_sequence[i] == 'O':
                    # Sometimes predict an entity where there isn't one
                    if random.random() < 0.3:
                        tag = random.choice(tags)
                        pred_sequence[i] = f'B-{tag}'
                else:
                    # Sometimes change the entity type or boundary
                    if random.random() < 0.3:
                        tag = random.choice(tags)
                        if pred_sequence[i].startswith('B-'):
                            pred_sequence[i] = f'B-{tag}'
                        elif pred_sequence[i].startswith('I-'):
                            pred_sequence[i] = f'I-{tag}'
                    elif random.random() < 0.3:
                        # Sometimes predict O instead of an entity
                        pred_sequence[i] = 'O'
        return pred_sequence
    
    true_sequences = []
    pred_sequences = []
    
    for _ in range(num_samples):
        length = random.randint(min_length, max_length)
        true_sequence = generate_sequence(length)
        pred_sequence = generate_prediction(true_sequence)
        true_sequences.append(true_sequence)
        pred_sequences.append(pred_sequence)
    
    return true_sequences, pred_sequences


def overall_report(true, pred):

    new_evaluator = NewEvaluator(true, pred, tags=['PER', 'ORG', 'LOC', 'DATE'], loader="list")
    print(new_evaluator.summary_report())

    print("-"*100)

    old_evaluator = OldEvaluator(true, pred, tags=['PER', 'ORG', 'LOC', 'DATE'], loader="list")
    results = old_evaluator.evaluate()[0]  # Get the first element which contains the overall results
    print(summary_report(results))


def entities_report(true, pred):

    new_evaluator = NewEvaluator(true, pred, tags=['PER', 'ORG', 'LOC', 'DATE'], loader="list")

    # entities - strict, exact, partial, ent_type
    print(new_evaluator.summary_report(mode="entities", scenario="strict"))
    print(new_evaluator.summary_report(mode="entities", scenario="exact"))
    print(new_evaluator.summary_report(mode="entities", scenario="partial"))
    print(new_evaluator.summary_report(mode="entities", scenario="ent_type"))

    print("-"*100)

    old_evaluator = OldEvaluator(true, pred, tags=['PER', 'ORG', 'LOC', 'DATE'], loader="list")
    _, results_agg_entities_type, _, _ = old_evaluator.evaluate()  # Get the second element which contains the entity-specific results
    print(summary_report(results_agg_entities_type, mode="entities", scenario="strict"))
    print(summary_report(results_agg_entities_type, mode="entities", scenario="exact"))
    print(summary_report(results_agg_entities_type, mode="entities", scenario="partial"))
    print(summary_report(results_agg_entities_type, mode="entities", scenario="ent_type"))


def indices_report_overall(true, pred):
    
    new_evaluator = NewEvaluator(true, pred, tags=['PER', 'LOC', 'DATE'], loader="list")
    print(new_evaluator.summary_report_indices(colors=True, mode="overall", scenario="strict"))
    print(new_evaluator.summary_report_indices(colors=True, mode="overall", scenario="exact"))
    print(new_evaluator.summary_report_indices(colors=True, mode="overall", scenario="partial"))
    print(new_evaluator.summary_report_indices(colors=True, mode="overall", scenario="ent_type"))

    old_evaluator = OldEvaluator(true, pred, tags=['LOC', 'PER', 'DATE'], loader="list")
    _, _, result_indices, _ = old_evaluator.evaluate()
    pred_dict = list_to_dict_format(pred)   # convert predictions to dictionary format for reporting
    print(summary_report_overall_indices(evaluation_indices=result_indices, error_schema='strict', preds=pred_dict))
    print(summary_report_overall_indices(evaluation_indices=result_indices, error_schema='exact', preds=pred_dict))
    print(summary_report_overall_indices(evaluation_indices=result_indices, error_schema='partial', preds=pred_dict))
    print(summary_report_overall_indices(evaluation_indices=result_indices, error_schema='ent_type', preds=pred_dict))


def indices_report_entities(true, pred):

    new_evaluator = NewEvaluator(true, pred, tags=['PER', 'LOC', 'DATE'], loader="list")
    print(new_evaluator.summary_report_indices(colors=True, mode="entities", scenario="strict"))
    print(new_evaluator.summary_report_indices(colors=True, mode="entities", scenario="exact"))
    print(new_evaluator.summary_report_indices(colors=True, mode="entities", scenario="partial"))
    print(new_evaluator.summary_report_indices(colors=True, mode="entities", scenario="ent_type"))

    old_evaluator = OldEvaluator(true, pred, tags=['LOC', 'PER', 'DATE'], loader="list")
    _, _, _, result_indices_by_tag = old_evaluator.evaluate()
    pred_dict = list_to_dict_format(pred)   # convert predictions to dictionary format for reporting
    print(summary_report_ents_indices(evaluation_agg_indices=result_indices_by_tag, error_schema='strict', preds=pred_dict))
    print(summary_report_ents_indices(evaluation_agg_indices=result_indices_by_tag, error_schema='exact', preds=pred_dict))
    print(summary_report_ents_indices(evaluation_agg_indices=result_indices_by_tag, error_schema='partial', preds=pred_dict))
    print(summary_report_ents_indices(evaluation_agg_indices=result_indices_by_tag, error_schema='ent_type', preds=pred_dict))


if __name__ == "__main__":
    tags = ['PER', 'ORG', 'LOC', 'DATE']
    true, pred = generate_synthetic_data(tags, num_samples=10)

    overall_report(true, pred)
    print("\n\n" + "="*100 + "\n\n")
    entities_report(true, pred)
    print("\n\n" + "="*100 + "\n\n")
    indices_report_overall(true, pred)
    print("\n\n" + "="*100 + "\n\n")
    indices_report_entities(true, pred)
    print("\n\n" + "="*100 + "\n\n")


from typing import List, Dict, Any, Union, Optional
import csv
import io

from .entities import EvaluationResult, EvaluationIndices
from .strategies import (
    EvaluationStrategy,
    StrictEvaluation,
    PartialEvaluation,
    EntityTypeEvaluation,
    ExactEvaluation,
)
from .loaders import DataLoader, ConllLoader, ListLoader, DictLoader
from .entities import Entity


class Evaluator:
    """Main evaluator class for NER evaluation."""

    def __init__(self, true: Any, pred: Any, tags: List[str], loader: str = "default") -> None:
        """
        Initialize the evaluator.

        Args:
            true: True entities in any supported format
            pred: Predicted entities in any supported format
            tags: List of valid entity tags
            loader: Name of the loader to use
        """
        self.tags = tags
        self._setup_loaders()
        self._load_data(true, pred, loader)
        self._setup_evaluation_strategies()

    def _setup_loaders(self) -> None:
        """Setup available data loaders."""
        self.loaders: Dict[str, DataLoader] = {"conll": ConllLoader(), "list": ListLoader(), "dict": DictLoader()}

    def _setup_evaluation_strategies(self) -> None:
        """Setup evaluation strategies."""
        self.strategies: Dict[str, EvaluationStrategy] = {
            "strict": StrictEvaluation(),
            "partial": PartialEvaluation(),
            "ent_type": EntityTypeEvaluation(),
            "exact": ExactEvaluation(),
        }

    def _load_data(self, true: Any, pred: Any, loader: str) -> None:
        """Load the true and predicted data."""
        if loader == "default":
            # Try to infer the loader based on input type
            if isinstance(true, str):
                loader = "conll"
            elif isinstance(true, list) and true and isinstance(true[0], list):
                if isinstance(true[0][0], dict):
                    loader = "dict"
                else:
                    loader = "list"
            else:
                raise ValueError("Could not infer loader from input type")

        if loader not in self.loaders:
            raise ValueError(f"Unknown loader: {loader}")

        # For list loader, check document lengths before loading
        if loader == "list":
            if len(true) != len(pred):
                raise ValueError("Number of predicted documents does not equal true")

            # Check that each document has the same length
            for i, (true_doc, pred_doc) in enumerate(zip(true, pred)):
                if len(true_doc) != len(pred_doc):
                    raise ValueError(f"Document {i} has different lengths: true={len(true_doc)}, pred={len(pred_doc)}")

        self.true = self.loaders[loader].load(true)
        self.pred = self.loaders[loader].load(pred)

        if len(self.true) != len(self.pred):
            raise ValueError("Number of predicted documents does not equal true")

    def evaluate(self) -> Dict[str, Any]:
        """
        Run the evaluation.

        Returns:
            Dictionary containing evaluation results for each strategy and entity type
        """
        results = {}
        # Get unique tags that appear in either true or predicted data
        used_tags = set()  # type: ignore
        for doc in self.true:
            used_tags.update(e.label for e in doc)
        for doc in self.pred:
            used_tags.update(e.label for e in doc)
        # Only keep tags that are both used and in the allowed tags list
        used_tags = used_tags.intersection(set(self.tags))

        entity_results: Dict[str, Dict[str, EvaluationResult]] = {tag: {} for tag in used_tags}
        indices = {}
        entity_indices: Dict[str, Dict[str, EvaluationIndices]] = {tag: {} for tag in used_tags}

        # Evaluate each document
        for doc_idx, (true_doc, pred_doc) in enumerate(zip(self.true, self.pred)):
            # Filter entities by valid tags
            true_doc = [e for e in true_doc if e.label in self.tags]
            pred_doc = [e for e in pred_doc if e.label in self.tags]

            # Evaluate with each strategy
            for strategy_name, strategy in self.strategies.items():
                result, doc_indices = strategy.evaluate(true_doc, pred_doc, self.tags, doc_idx)

                # Update overall results
                if strategy_name not in results:
                    results[strategy_name] = result
                    indices[strategy_name] = doc_indices
                else:
                    self._merge_results(results[strategy_name], result)
                    self._merge_indices(indices[strategy_name], doc_indices)

                # Update entity-specific results
                for tag in used_tags:
                    # Filter entities for this specific tag
                    true_tag_doc = [e for e in true_doc if e.label == tag]
                    pred_tag_doc = [e for e in pred_doc if e.label == tag]

                    # Evaluate only entities of this tag
                    tag_result, tag_indices = strategy.evaluate(true_tag_doc, pred_tag_doc, [tag], doc_idx)

                    if tag not in entity_results:
                        entity_results[tag] = {}
                        entity_indices[tag] = {}
                    if strategy_name not in entity_results[tag]:
                        entity_results[tag][strategy_name] = tag_result
                        entity_indices[tag][strategy_name] = tag_indices
                    else:
                        self._merge_results(entity_results[tag][strategy_name], tag_result)
                        self._merge_indices(entity_indices[tag][strategy_name], tag_indices)

        return {
            "overall": results,
            "entities": entity_results,
            "overall_indices": indices,
            "entity_indices": entity_indices,
        }

    @staticmethod
    def _merge_results(target: EvaluationResult, source: EvaluationResult) -> None:
        """Merge two evaluation results."""
        target.correct += source.correct
        target.incorrect += source.incorrect
        target.partial += source.partial
        target.missed += source.missed
        target.spurious += source.spurious
        target.compute_metrics()

    @staticmethod
    def _merge_indices(target: EvaluationIndices, source: EvaluationIndices) -> None:
        """Merge two evaluation indices."""
        target.correct_indices.extend(source.correct_indices)
        target.incorrect_indices.extend(source.incorrect_indices)
        target.partial_indices.extend(source.partial_indices)
        target.missed_indices.extend(source.missed_indices)
        target.spurious_indices.extend(source.spurious_indices)

    def results_to_csv(
        self, mode: str = "overall", scenario: str = "strict", file_path: Optional[str] = None
    ) -> Union[str, None]:
        """
        Convert results to CSV format.

        Args:
            mode: Either 'overall' for overall metrics or 'entities' for per-entity metrics
            scenario: The scenario to report on (only used when mode is 'entities')
            file_path: Optional path to save CSV file. If None, returns CSV as string

        Returns:
            CSV content as string if file_path is None, otherwise None (saves to file)
        """
        valid_modes = {"overall", "entities"}
        valid_scenarios = {"strict", "ent_type", "partial", "exact"}

        if mode not in valid_modes:
            raise ValueError(f"Invalid mode: must be one of {valid_modes}")

        if mode == "entities" and scenario not in valid_scenarios:
            raise ValueError(f"Invalid scenario: must be one of {valid_scenarios}")

        results = self.evaluate()

        if mode == "overall":
            # For overall mode, include all scenarios
            csv_data = [
                ["Strategy", "Correct", "Incorrect", "Partial", "Missed", "Spurious", "Precision", "Recall", "F1-Score"]
            ]
            results_data = results["overall"]
            for strategy_name, strategy_result in results_data.items():
                csv_data.append(
                    [
                        strategy_name,
                        strategy_result.correct,
                        strategy_result.incorrect,
                        strategy_result.partial,
                        strategy_result.missed,
                        strategy_result.spurious,
                        strategy_result.precision,
                        strategy_result.recall,
                        strategy_result.f1,
                    ]
                )
        else:
            csv_data = [
                ["Entity", "Correct", "Incorrect", "Partial", "Missed", "Spurious", "Precision", "Recall", "F1-Score"]
            ]
            results_data = results["entities"]
            for entity_type, entity_results in results_data.items():
                if scenario in entity_results:
                    strategy_result = entity_results[scenario]
                    csv_data.append(
                        [
                            entity_type,
                            strategy_result.correct,
                            strategy_result.incorrect,
                            strategy_result.partial,
                            strategy_result.missed,
                            strategy_result.spurious,
                            strategy_result.precision,
                            strategy_result.recall,
                            strategy_result.f1,
                        ]
                    )

        if file_path:
            with open(file_path, "w", newline="", encoding="utf-8") as csvfile:
                writer = csv.writer(csvfile)
                writer.writerows(csv_data)
            return None

        output = io.StringIO()
        writer = csv.writer(output)
        writer.writerows(csv_data)
        return output.getvalue()

    def summary_report(self, mode: str = "overall", scenario: str = "strict", digits: int = 2) -> str:
        """
        Generate a summary report of the evaluation results.

        Args:
            mode: Either 'overall' for overall metrics or 'entities' for per-entity metrics.
            scenario: The scenario to report on. Only used when mode is 'entities'.
                      Must be one of:
                        - 'strict' exact boundary surface string match and entity type;
                        - 'exact': exact boundary match over the surface string and entity type;
                        - 'partial': partial boundary match over the surface string, regardless of the type;
                        - 'ent_type': exact boundary match over the surface string, regardless of the type;
            digits: The number of digits to round the results to.

        Returns:
            A string containing the summary report.

        Raises:
            ValueError: If the scenario or mode is invalid.
        """
        valid_scenarios = {"strict", "ent_type", "partial", "exact"}
        valid_modes = {"overall", "entities"}

        if mode not in valid_modes:
            raise ValueError(f"Invalid mode: must be one of {valid_modes}")

        if mode == "entities" and scenario not in valid_scenarios:
            raise ValueError(f"Invalid scenario: must be one of {valid_scenarios}")

        headers = ["correct", "incorrect", "partial", "missed", "spurious", "precision", "recall", "f1-score"]
        rows = [headers]

        results = self.evaluate()
        if mode == "overall":
            # Process overall results - show all scenarios
            results_data = results["overall"]
            for eval_schema in sorted(valid_scenarios):  # Sort to ensure consistent order
                if eval_schema not in results_data:
                    continue
                results_schema = results_data[eval_schema]
                rows.append(
                    [
                        eval_schema,
                        results_schema.correct,
                        results_schema.incorrect,
                        results_schema.partial,
                        results_schema.missed,
                        results_schema.spurious,
                        results_schema.precision,
                        results_schema.recall,
                        results_schema.f1,
                    ]
                )
        else:
            # Process entity-specific results for the specified scenario only
            results_data = results["entities"]
            target_names = sorted(results_data.keys())
            for ent_type in target_names:
                if scenario not in results_data[ent_type]:
                    continue  # Skip if scenario not available for this entity type

                results_ent = results_data[ent_type][scenario]
                rows.append(
                    [
                        ent_type,
                        results_ent.correct,
                        results_ent.incorrect,
                        results_ent.partial,
                        results_ent.missed,
                        results_ent.spurious,
                        results_ent.precision,
                        results_ent.recall,
                        results_ent.f1,
                    ]
                )

        # Format the report
        name_width = max(len(str(row[0])) for row in rows)
        width = max(name_width, digits)
        head_fmt = "{:>{width}s} " + " {:>11}" * len(headers)
        report = f"Scenario: {scenario if mode == 'entities' else 'all'}\n\n" + head_fmt.format(
            "", *headers, width=width
        )
        report += "\n\n"
        row_fmt = "{:>{width}s} " + " {:>11}" * 5 + " {:>11.{digits}f}" * 3 + "\n"

        for row in rows[1:]:
            report += row_fmt.format(*row, width=width, digits=digits)

        return report

    def summary_report_indices(  # pylint: disable=too-many-branches
        self, mode: str = "overall", scenario: str = "strict", colors: bool = False
    ) -> str:
        """
        Generate a summary report of the evaluation indices.

        Args:
            mode: Either 'overall' for overall metrics or 'entities' for per-entity metrics.
            scenario: The scenario to report on. Must be one of: 'strict', 'ent_type', 'partial', 'exact'.
                     Only used when mode is 'entities'. Defaults to 'strict'.
            colors: Whether to use colors in the output. Defaults to False.

        Returns:
            A string containing the summary report of indices.

        Raises:
            ValueError: If the scenario or mode is invalid.
        """
        valid_scenarios = {"strict", "ent_type", "partial", "exact"}
        valid_modes = {"overall", "entities"}

        if mode not in valid_modes:
            raise ValueError(f"Invalid mode: must be one of {valid_modes}")

        if mode == "entities" and scenario not in valid_scenarios:
            raise ValueError(f"Invalid scenario: must be one of {valid_scenarios}")

        # ANSI color codes
        COLORS = {
            "reset": "\033[0m",
            "bold": "\033[1m",
            "red": "\033[91m",
            "green": "\033[92m",
            "yellow": "\033[93m",
            "blue": "\033[94m",
            "magenta": "\033[95m",
            "cyan": "\033[96m",
            "white": "\033[97m",
        }

        def colorize(text: str, color: str) -> str:
            """Helper function to colorize text if colors are enabled."""
            if colors:
                return f"{COLORS[color]}{text}{COLORS['reset']}"
            return text

        def get_prediction_info(pred: Union[Entity, str]) -> str:
            """Helper function to get prediction info based on pred type."""
            if isinstance(pred, Entity):
                return f"Label={pred.label}, Start={pred.start}, End={pred.end}"
            # String (BIO tag)
            return f"Tag={pred}"

        results = self.evaluate()
        report = ""

        # Create headers for the table
        headers = ["Category", "Instance", "Entity", "Details"]
        header_fmt = "{:<20} {:<10} {:<8} {:<25}"
        row_fmt = "{:<20} {:<10} {:<8} {:<10}"

        if mode == "overall":
            # Get the indices from the overall results
            indices_data = results["overall_indices"][scenario]
            report += f"\n{colorize('Indices for error schema', 'bold')} '{colorize(scenario, 'cyan')}':\n\n"
            report += colorize(header_fmt.format(*headers), "bold") + "\n"
            report += colorize("-" * 78, "white") + "\n"

            for category, indices in indices_data.__dict__.items():
                if not category.endswith("_indices"):
                    continue
                category_name = category.replace("_indices", "").replace("_", " ").capitalize()

                # Color mapping for categories
                category_colors = {
                    "Correct": "green",
                    "Incorrect": "red",
                    "Partial": "yellow",
                    "Missed": "magenta",
                    "Spurious": "blue",
                }

                if indices:
                    for instance_index, entity_index in indices:
                        if self.pred != [[]]:
                            pred = self.pred[instance_index][entity_index]
                            prediction_info = get_prediction_info(pred)
                            report += (
                                row_fmt.format(
                                    colorize(category_name, category_colors.get(category_name, "white")),
                                    f"{instance_index}",
                                    f"{entity_index}",
                                    prediction_info,
                                )
                                + "\n"
                            )
                        else:
                            report += (
                                row_fmt.format(
                                    colorize(category_name, category_colors.get(category_name, "white")),
                                    f"{instance_index}",
                                    f"{entity_index}",
                                    "No prediction info",
                                )
                                + "\n"
                            )
                else:
                    report += (
                        row_fmt.format(
                            colorize(category_name, category_colors.get(category_name, "white")), "-", "-", "None"
                        )
                        + "\n"
                    )
        else:
            # Get the indices from the entity-specific results
            for entity_type, entity_results in results["entity_indices"].items():
                report += f"\n{colorize('Entity Type', 'bold')}: {colorize(entity_type, 'cyan')}\n"
                report += f"{colorize('Error Schema', 'bold')}: '{colorize(scenario, 'cyan')}'\n\n"
                report += colorize(header_fmt.format(*headers), "bold") + "\n"
                report += colorize("-" * 78, "white") + "\n"

                error_data = entity_results[scenario]
                for category, indices in error_data.__dict__.items():
                    if not category.endswith("_indices"):
                        continue
                    category_name = category.replace("_indices", "").replace("_", " ").capitalize()

                    # Color mapping for categories
                    category_colors = {
                        "Correct": "green",
                        "Incorrect": "red",
                        "Partial": "yellow",
                        "Missed": "magenta",
                        "Spurious": "blue",
                    }

                    if indices:
                        for instance_index, entity_index in indices:
                            if self.pred != [[]]:
                                pred = self.pred[instance_index][entity_index]
                                prediction_info = get_prediction_info(pred)
                                report += (
                                    row_fmt.format(
                                        colorize(category_name, category_colors.get(category_name, "white")),
                                        f"{instance_index}",
                                        f"{entity_index}",
                                        prediction_info,
                                    )
                                    + "\n"
                                )
                            else:
                                report += (
                                    row_fmt.format(
                                        colorize(category_name, category_colors.get(category_name, "white")),
                                        f"{instance_index}",
                                        f"{entity_index}",
                                        "No prediction info",
                                    )
                                    + "\n"
                                )
                    else:
                        report += (
                            row_fmt.format(
                                colorize(category_name, category_colors.get(category_name, "white")), "-", "-", "None"
                            )
                            + "\n"
                        )

        return report



import csv
import io
import pytest
from nervaluate.evaluator import Evaluator


@pytest.fixture
def sample_data():
    true = [
        ["O", "B-PER", "O", "B-ORG", "I-ORG", "B-LOC"],
        ["O", "B-PER", "O", "B-ORG"],
    ]

    pred = [
        ["O", "B-PER", "O", "B-ORG", "O", "B-PER"],
        ["O", "B-PER", "O", "B-LOC"],
    ]

    return true, pred


def test_evaluator_initialization(sample_data):
    """Test evaluator initialization."""
    true, pred = sample_data
    evaluator = Evaluator(true, pred, ["PER", "ORG", "LOC"], loader="list")

    assert len(evaluator.true) == 2
    assert len(evaluator.pred) == 2
    assert evaluator.tags == ["PER", "ORG", "LOC"]


def test_evaluator_evaluation(sample_data):
    """Test evaluation process."""
    true, pred = sample_data
    evaluator = Evaluator(true, pred, ["PER", "ORG", "LOC"], loader="list")
    results = evaluator.evaluate()

    # Check that we have results for all strategies
    assert "overall" in results
    assert "entities" in results
    assert "strict" in results["overall"]
    assert "partial" in results["overall"]
    assert "ent_type" in results["overall"]

    # Check that we have results for each entity type
    for entity in ["PER", "ORG", "LOC"]:
        assert entity in results["entities"]
        assert "strict" in results["entities"][entity]
        assert "partial" in results["entities"][entity]
        assert "ent_type" in results["entities"][entity]


def test_evaluator_with_invalid_tags(sample_data):
    """Test evaluator with invalid tags."""
    true, pred = sample_data
    evaluator = Evaluator(true, pred, ["INVALID"], loader="list")
    results = evaluator.evaluate()

    for strategy in ["strict", "partial", "ent_type"]:
        assert results["overall"][strategy].correct == 0
        assert results["overall"][strategy].incorrect == 0
        assert results["overall"][strategy].partial == 0
        assert results["overall"][strategy].missed == 0
        assert results["overall"][strategy].spurious == 0


def test_evaluator_different_document_lengths():
    """Test that Evaluator raises ValueError when documents have different lengths."""
    true = [
        ["O", "B-PER", "I-PER", "O", "O", "O", "B-ORG", "I-ORG"],  # 8 tokens
        ["O", "B-LOC", "B-PER", "I-PER", "O", "O", "B-DATE"],  # 7 tokens
    ]
    pred = [
        ["O", "B-PER", "I-PER", "O", "O", "O", "B-ORG", "I-ORG"],  # 8 tokens
        ["O", "B-LOC", "I-LOC", "O", "B-PER", "I-PER", "O", "B-DATE", "I-DATE", "O"],  # 10 tokens
    ]
    tags = ["PER", "ORG", "LOC", "DATE"]

    # Test that ValueError is raised
    with pytest.raises(ValueError, match="Document 1 has different lengths: true=7, pred=10"):
        evaluator = Evaluator(true=true, pred=pred, tags=tags, loader="list")
        evaluator.evaluate()


def test_results_to_csv(sample_data, tmp_path):

    true, pred = sample_data
    evaluator = Evaluator(true, pred, ["PER", "ORG", "LOC"], loader="list")

    overall_csv_str = evaluator.results_to_csv(mode="overall")
    assert isinstance(overall_csv_str, str)

    csv_reader = csv.reader(io.StringIO(overall_csv_str))
    overall_csv = list(csv_reader)

    assert len(overall_csv) > 1  # should have header + at least one row
    assert overall_csv[0] == [
        "Strategy",
        "Correct",
        "Incorrect",
        "Partial",
        "Missed",
        "Spurious",
        "Precision",
        "Recall",
        "F1-Score",
    ]

    # check that all strategies are present
    strategies = {row[0] for row in overall_csv[1:]}
    assert strategies == {"strict", "partial", "ent_type", "exact"}

    # test entities mode - return as string
    entities_csv_str = evaluator.results_to_csv(mode="entities", scenario="strict")
    assert isinstance(entities_csv_str, str)

    # parse CSV string to check content
    csv_reader = csv.reader(io.StringIO(entities_csv_str))
    entities_csv = list(csv_reader)

    assert len(entities_csv) > 1  # should have header + at least one row
    assert entities_csv[0] == [
        "Entity",
        "Correct",
        "Incorrect",
        "Partial",
        "Missed",
        "Spurious",
        "Precision",
        "Recall",
        "F1-Score",
    ]

    # check that all entity types are present
    entity_types = {row[0] for row in entities_csv[1:]}
    assert entity_types == {"PER", "ORG", "LOC"}

    # test file saving - overall mode
    overall_file = tmp_path / "overall_results.csv"
    result = evaluator.results_to_csv(mode="overall", file_path=str(overall_file))
    assert result is None  # Should return None when saving to file
    assert overall_file.exists()

    # verify file content
    with open(overall_file, "r", encoding="utf-8") as f:
        saved_csv = list(csv.reader(f))
    assert len(saved_csv) > 1
    assert saved_csv[0][0] == "Strategy"

    # test file saving - entities mode
    entities_file = tmp_path / "entities_results.csv"
    result = evaluator.results_to_csv(mode="entities", scenario="partial", file_path=str(entities_file))
    assert result is None  # Should return None when saving to file
    assert entities_file.exists()

    # verify file content
    with open(entities_file, "r", encoding="utf-8") as f:
        saved_csv = list(csv.reader(f))
    assert len(saved_csv) > 1
    assert saved_csv[0][0] == "Entity"

    # test invalid mode
    with pytest.raises(ValueError, match="Invalid mode: must be one of"):
        evaluator.results_to_csv(mode="invalid")

    # test invalid scenario for entities mode
    with pytest.raises(ValueError, match="Invalid scenario: must be one of"):
        evaluator.results_to_csv(mode="entities", scenario="invalid")



from nervaluate.entities import Entity, EvaluationResult


def test_entity_equality():
    """Test Entity equality comparison."""
    entity1 = Entity(label="PER", start=0, end=1)
    entity2 = Entity(label="PER", start=0, end=1)
    entity3 = Entity(label="ORG", start=0, end=1)

    assert entity1 == entity2
    assert entity1 != entity3
    assert entity1 != "not an entity"


def test_entity_hash():
    """Test Entity hashing."""
    entity1 = Entity(label="PER", start=0, end=1)
    entity2 = Entity(label="PER", start=0, end=1)
    entity3 = Entity(label="ORG", start=0, end=1)

    assert hash(entity1) == hash(entity2)
    assert hash(entity1) != hash(entity3)


def test_evaluation_result_compute_metrics():
    """Test computation of evaluation metrics."""
    result = EvaluationResult(correct=5, incorrect=2, partial=1, missed=1, spurious=1)

    # Test strict metrics
    result.compute_metrics(partial_or_type=False)
    assert result.precision == 5 / 9  # 5/(5+2+1+1)
    assert result.recall == 5 / (5 + 2 + 1 + 1)

    # Test partial metrics
    result.compute_metrics(partial_or_type=True)
    assert result.precision == 5.5 / 9  # (5+0.5*1)/(5+2+1+1)
    assert result.recall == (5 + 0.5 * 1) / (5 + 2 + 1 + 1)


def test_evaluation_result_zero_cases():
    """Test evaluation metrics with zero values."""
    result = EvaluationResult()
    result.compute_metrics()
    assert result.precision == 0
    assert result.recall == 0
    assert result.f1 == 0



from nervaluate import (
    collect_named_entities,
    conll_to_spans,
    list_to_spans,
    split_list,
)


def test_list_to_spans():
    before = [
        ["O", "B-LOC", "I-LOC", "B-LOC", "I-LOC", "O"],
        ["O", "B-GPE", "I-GPE", "B-GPE", "I-GPE", "O"],
    ]

    expected = [
        [
            {"label": "LOC", "start": 1, "end": 2},
            {"label": "LOC", "start": 3, "end": 4},
        ],
        [
            {"label": "GPE", "start": 1, "end": 2},
            {"label": "GPE", "start": 3, "end": 4},
        ],
    ]

    result = list_to_spans(before)

    assert result == expected


def test_list_to_spans_1():
    before = [
        ["O", "O", "O", "O", "O", "O"],
        ["O", "O", "B-ORG", "I-ORG", "O", "O"],
        ["O", "O", "B-MISC", "I-MISC", "O", "O"],
    ]

    expected = [
        [],
        [{"label": "ORG", "start": 2, "end": 3}],
        [{"label": "MISC", "start": 2, "end": 3}],
    ]

    actual = list_to_spans(before)

    assert actual == expected


def test_conll_to_spans():
    before = (
        ",\tO\n"
        "Davos\tB-PER\n"
        "2018\tO\n"
        ":\tO\n"
        "Soros\tB-PER\n"
        "accuses\tO\n"
        "Trump\tB-PER\n"
        "of\tO\n"
        "wanting\tO\n"
        "\n"
        "foo\tO\n"
    )

    after = [
        [
            {"label": "PER", "start": 1, "end": 1},
            {"label": "PER", "start": 4, "end": 4},
            {"label": "PER", "start": 6, "end": 6},
        ],
        [],
    ]

    out = conll_to_spans(before)

    assert after == out


def test_conll_to_spans_1():
    before = (
        "word\tO\nword\tO\nword\tO\nword\tO\nword\tO\nword\tO\n\n"
        "word\tO\nword\tO\nword\tB-ORG\nword\tI-ORG\nword\tO\nword\tO\n\n"
        "word\tO\nword\tO\nword\tB-MISC\nword\tI-MISC\nword\tO\nword\tO\n"
    )

    expected = [
        [],
        [{"label": "ORG", "start": 2, "end": 3}],
        [{"label": "MISC", "start": 2, "end": 3}],
    ]

    actual = conll_to_spans(before)

    assert actual == expected


def test_split_list():
    before = ["aa", "bb", "cc", "", "dd", "ee", "ff"]
    expected = [["aa", "bb", "cc"], ["dd", "ee", "ff"]]
    out = split_list(before)

    assert expected == out


def test_collect_named_entities_same_type_in_sequence():
    tags = ["O", "B-LOC", "I-LOC", "B-LOC", "I-LOC", "O"]
    result = collect_named_entities(tags)
    expected = [
        {"label": "LOC", "start": 1, "end": 2},
        {"label": "LOC", "start": 3, "end": 4},
    ]
    assert result == expected


def test_collect_named_entities_sequence_has_only_one_entity():
    tags = ["B-LOC", "I-LOC"]
    result = collect_named_entities(tags)
    expected = [{"label": "LOC", "start": 0, "end": 1}]
    assert result == expected


def test_collect_named_entities_entity_goes_until_last_token():
    tags = ["O", "B-LOC", "I-LOC", "B-LOC", "I-LOC"]
    result = collect_named_entities(tags)
    expected = [
        {"label": "LOC", "start": 1, "end": 2},
        {"label": "LOC", "start": 3, "end": 4},
    ]
    assert result == expected


def test_collect_named_entities_no_entity():
    tags = ["O", "O", "O", "O", "O"]
    result = collect_named_entities(tags)
    expected = []
    assert result == expected



import nltk
import sklearn_crfsuite
from sklearn.metrics import classification_report

from nervaluate import Evaluator, collect_named_entities, summary_report_ent, summary_report_overall


def word2features(sent, i):
    word = sent[i][0]
    postag = sent[i][1]

    features = {
        "bias": 1.0,
        "word.lower()": word.lower(),
        "word[-3:]": word[-3:],
        "word[-2:]": word[-2:],
        "word.isupper()": word.isupper(),
        "word.istitle()": word.istitle(),
        "word.isdigit()": word.isdigit(),
        "postag": postag,
        "postag[:2]": postag[:2],
    }
    if i > 0:
        word1 = sent[i - 1][0]
        postag1 = sent[i - 1][1]
        features.update(
            {
                "-1:word.lower()": word1.lower(),
                "-1:word.istitle()": word1.istitle(),
                "-1:word.isupper()": word1.isupper(),
                "-1:postag": postag1,
                "-1:postag[:2]": postag1[:2],
            }
        )
    else:
        features["BOS"] = True

    if i < len(sent) - 1:
        word1 = sent[i + 1][0]
        postag1 = sent[i + 1][1]
        features.update(
            {
                "+1:word.lower()": word1.lower(),
                "+1:word.istitle()": word1.istitle(),
                "+1:word.isupper()": word1.isupper(),
                "+1:postag": postag1,
                "+1:postag[:2]": postag1[:2],
            }
        )
    else:
        features["EOS"] = True

    return features


def sent2features(sent):
    return [word2features(sent, i) for i in range(len(sent))]


def sent2labels(sent):
    return [label for token, postag, label in sent]


def sent2tokens(sent):
    return [token for token, postag, label in sent]


def main():
    print("Loading CoNLL 2002 NER Spanish data")
    nltk.corpus.conll2002.fileids()
    train_sents = list(nltk.corpus.conll2002.iob_sents("esp.train"))
    test_sents = list(nltk.corpus.conll2002.iob_sents("esp.testb"))

    x_train = [sent2features(s) for s in train_sents]
    y_train = [sent2labels(s) for s in train_sents]

    x_test = [sent2features(s) for s in test_sents]
    y_test = [sent2labels(s) for s in test_sents]

    print("Train a CRF on the CoNLL 2002 NER Spanish data")
    crf = sklearn_crfsuite.CRF(algorithm="lbfgs", c1=0.1, c2=0.1, max_iterations=10, all_possible_transitions=True)
    try:
        crf.fit(x_train, y_train)
    except AttributeError:
        pass

    y_pred = crf.predict(x_test)
    labels = list(crf.classes_)
    labels.remove("O")  # remove 'O' label from evaluation
    sorted_labels = sorted(labels, key=lambda name: (name[1:], name[0]))  # group B- and I- results
    y_test_flat = [y for msg in y_test for y in msg]
    y_pred_flat = [y for msg in y_pred for y in msg]
    print(classification_report(y_test_flat, y_pred_flat, labels=sorted_labels))

    test_sents_labels = []
    for sentence in test_sents:
        sentence = [token[2] for token in sentence]
        test_sents_labels.append(sentence)

    pred_collected = [collect_named_entities(msg) for msg in y_pred]
    test_collected = [collect_named_entities(msg) for msg in y_test]

    evaluator = Evaluator(test_collected, pred_collected, ["LOC", "MISC", "PER", "ORG"])
    results, results_agg = evaluator.evaluate()

    print("\n\nOverall")
    print(summary_report_overall(results))
    print("\n\n'Strict'")
    print(summary_report_ent(results_agg, scenario="strict"))
    print("\n\n'Ent_Type'")
    print(summary_report_ent(results_agg, scenario="ent_type"))
    print("\n\n'Partial'")
    print(summary_report_ent(results_agg, scenario="partial"))
    print("\n\n'Exact'")
    print(summary_report_ent(results_agg, scenario="exact"))


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from typing import List, Tuple

from .entities import Entity, EvaluationResult, EvaluationIndices


class EvaluationStrategy(ABC):
    """Abstract base class for evaluation strategies."""

    @abstractmethod
    def evaluate(
        self, true_entities: List[Entity], pred_entities: List[Entity], tags: List[str], instance_index: int = 0
    ) -> Tuple[EvaluationResult, EvaluationIndices]:
        """Evaluate the predicted entities against the true entities."""


class StrictEvaluation(EvaluationStrategy):
    """
    Strict evaluation strategy - entities must match exactly.

    If there's a predicted entity that perfectly matches a true entity and they have the same label
    we mark it as correct.
    If there's a predicted entity that doesn't perfectly match any true entity, we mark it as spurious.
    If there's a true entity that doesn't perfecly match any predicted entity, we mark it as missed.
    All other cases are marked as incorrect.
    """

    def evaluate(
        self, true_entities: List[Entity], pred_entities: List[Entity], tags: List[str], instance_index: int = 0
    ) -> Tuple[EvaluationResult, EvaluationIndices]:
        """
        Evaluate the predicted entities against the true entities using strict matching.
        """
        result = EvaluationResult()
        indices = EvaluationIndices()
        matched_true = set()

        for pred_idx, pred in enumerate(pred_entities):
            found_match = False
            found_incorrect = False

            for true_idx, true in enumerate(true_entities):
                if true_idx in matched_true:
                    continue

                # Check for perfect match (same boundaries and label)
                if pred.label == true.label and pred.start == true.start and pred.end == true.end:
                    result.correct += 1
                    indices.correct_indices.append((instance_index, pred_idx))
                    matched_true.add(true_idx)
                    found_match = True
                    break
                # Check for any overlap
                if pred.start <= true.end and pred.end >= true.start:
                    result.incorrect += 1
                    indices.incorrect_indices.append((instance_index, pred_idx))
                    matched_true.add(true_idx)
                    found_incorrect = True
                    break

            if not found_match and not found_incorrect:
                result.spurious += 1
                indices.spurious_indices.append((instance_index, pred_idx))

        for true_idx, true in enumerate(true_entities):
            if true_idx not in matched_true:
                result.missed += 1
                indices.missed_indices.append((instance_index, true_idx))

        result.compute_metrics()
        return result, indices


class PartialEvaluation(EvaluationStrategy):
    """
    Partial evaluation strategy - allows for partial matches.

    If there's a predicted entity that perfectly matches a true entity, we mark it as correct.
    If there's a predicted entity that has some minimum overlap with a true entity we mark it as partial.
    If there's a predicted entity that doesn't match any true entity, we mark it as spurious.
    If there's a true entity that doesn't match any predicted entity, we mark it as missed.

    There's never entity type/label checking in this strategy, and there's never an entity marked as incorrect.
    """

    def evaluate(
        self, true_entities: List[Entity], pred_entities: List[Entity], tags: List[str], instance_index: int = 0
    ) -> Tuple[EvaluationResult, EvaluationIndices]:
        result = EvaluationResult()
        indices = EvaluationIndices()
        matched_true = set()

        for pred_idx, pred in enumerate(pred_entities):
            found_match = False

            for true_idx, true in enumerate(true_entities):
                if true_idx in matched_true:
                    continue

                # Check for overlap
                if pred.start <= true.end and pred.end >= true.start:
                    if pred.start == true.start and pred.end == true.end:
                        result.correct += 1
                        indices.correct_indices.append((instance_index, pred_idx))
                    else:
                        result.partial += 1
                        indices.partial_indices.append((instance_index, pred_idx))
                    matched_true.add(true_idx)
                    found_match = True
                    break

            if not found_match:
                result.spurious += 1
                indices.spurious_indices.append((instance_index, pred_idx))

        for true_idx, true in enumerate(true_entities):
            if true_idx not in matched_true:
                result.missed += 1
                indices.missed_indices.append((instance_index, true_idx))

        result.compute_metrics(partial_or_type=True)
        return result, indices


class EntityTypeEvaluation(EvaluationStrategy):
    """
    Entity type evaluation strategy - only checks entity types.

    In in strategy, we check for overlap between the predicted entity and the true entity.

    If there's a predicted entity that perfectly matches or only some minimum overlap with a
    true entity, and the same label, we mark it as correct.
    If there's a predicted entity that has some minimum overlap or perfectly matches but has
    the wrong label we mark it as inccorrect.
    If there's a predicted entity that doesn't match any true entity, we mark it as spurious.
    If there's a true entity that doesn't match any predicted entity, we mark it as missed.

    # ToDo: define a minimum overlap threshold - see: https://github.com/MantisAI/nervaluate/pull/83
    """

    def evaluate(
        self, true_entities: List[Entity], pred_entities: List[Entity], tags: List[str], instance_index: int = 0
    ) -> Tuple[EvaluationResult, EvaluationIndices]:
        result = EvaluationResult()
        indices = EvaluationIndices()
        matched_true = set()

        for pred_idx, pred in enumerate(pred_entities):
            found_match = False
            found_incorrect = False

            for true_idx, true in enumerate(true_entities):
                if true_idx in matched_true:
                    continue

                # Check for any overlap (perfect or minimum)
                if pred.start <= true.end and pred.end >= true.start:
                    if pred.label == true.label:
                        result.correct += 1
                        indices.correct_indices.append((instance_index, pred_idx))
                        matched_true.add(true_idx)
                        found_match = True
                    else:
                        result.incorrect += 1
                        indices.incorrect_indices.append((instance_index, pred_idx))
                        matched_true.add(true_idx)
                        found_incorrect = True
                    break

            if not found_match and not found_incorrect:
                result.spurious += 1
                indices.spurious_indices.append((instance_index, pred_idx))

        for true_idx, true in enumerate(true_entities):
            if true_idx not in matched_true:
                result.missed += 1
                indices.missed_indices.append((instance_index, true_idx))

        result.compute_metrics(partial_or_type=True)
        return result, indices


class ExactEvaluation(EvaluationStrategy):
    """
    Exact evaluation strategy - exact boundary match over the surface string, regardless of the type.

    If there's a predicted entity that perfectly matches a true entity, regardless of the label, we mark it as correct.
    If there's a predicted entity that has only some minimum overlap with a true entity, we mark it as incorrect.
    If there's a predicted entity that doesn't match any true entity, we mark it as spurious.
    If there's a true entity that doesn't match any predicted entity, we mark it as missed.
    """

    def evaluate(
        self, true_entities: List[Entity], pred_entities: List[Entity], tags: List[str], instance_index: int = 0
    ) -> Tuple[EvaluationResult, EvaluationIndices]:
        """
        Evaluate the predicted entities against the true entities using exact boundary matching.
        Entity type is not considered in the matching.
        """
        result = EvaluationResult()
        indices = EvaluationIndices()
        matched_true = set()

        for pred_idx, pred in enumerate(pred_entities):
            found_match = False
            found_incorrect = False

            for true_idx, true in enumerate(true_entities):
                if true_idx in matched_true:
                    continue

                # Check for exact boundary match (regardless of label)
                if pred.start == true.start and pred.end == true.end:
                    result.correct += 1
                    indices.correct_indices.append((instance_index, pred_idx))
                    matched_true.add(true_idx)
                    found_match = True
                    break
                # Check for any overlap
                if pred.start <= true.end and pred.end >= true.start:
                    result.incorrect += 1
                    indices.incorrect_indices.append((instance_index, pred_idx))
                    matched_true.add(true_idx)
                    found_incorrect = True
                    break

            if not found_match and not found_incorrect:
                result.spurious += 1
                indices.spurious_indices.append((instance_index, pred_idx))

        for true_idx, true in enumerate(true_entities):
            if true_idx not in matched_true:
                result.missed += 1
                indices.missed_indices.append((instance_index, true_idx))

        result.compute_metrics()
        return result, indices



import pytest
from nervaluate.entities import Entity
from nervaluate.strategies import EntityTypeEvaluation, ExactEvaluation, PartialEvaluation, StrictEvaluation


def create_entities_from_bio(bio_tags):
    """Helper function to create entities from BIO tags."""
    entities = []
    current_entity = None

    for i, tag in enumerate(bio_tags):
        if tag == "O":
            continue

        if tag.startswith("B-"):
            if current_entity:
                entities.append(current_entity)
            current_entity = Entity(tag[2:], i, i + 1)
        elif tag.startswith("I-"):
            if current_entity:
                current_entity.end = i + 1
            else:
                # Handle case where I- tag appears without B-
                current_entity = Entity(tag[2:], i, i + 1)

    if current_entity:
        entities.append(current_entity)

    return entities


@pytest.fixture
def base_sequence():
    """Base sequence: 'The John Smith who works at Google Inc'"""
    return ["O", "B-PER", "I-PER", "O", "O", "O", "B-ORG", "I-ORG"]


class TestStrictEvaluation:
    """Test cases for strict evaluation strategy."""

    def test_perfect_match(self, base_sequence):
        """Test case: Perfect match of all entities."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(base_sequence)

        evaluator = StrictEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG"])

        assert result.correct == 2
        assert result.incorrect == 0
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0), (0, 1)]
        assert result_indices.incorrect_indices == []
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == []

    def test_missed_entity(self, base_sequence):
        """Test case: One entity is missed in prediction."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "O", "O"])

        evaluator = StrictEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG"])

        assert result.correct == 1
        assert result.incorrect == 0
        assert result.partial == 0
        assert result.missed == 1
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0)]
        assert result_indices.incorrect_indices == []
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == [(0, 1)]
        assert result_indices.spurious_indices == []

    def test_wrong_label(self, base_sequence):
        """Test case: Entity with wrong label."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "B-LOC", "I-LOC"])

        evaluator = StrictEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 1
        assert result.incorrect == 1
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0)]
        assert result_indices.incorrect_indices == [(0, 1)]
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == []

    def test_wrong_boundary(self, base_sequence):
        """Test case: Entity with wrong boundary."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "B-LOC", "O"])

        evaluator = StrictEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 1
        assert result.incorrect == 1
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0)]
        assert result_indices.incorrect_indices == [(0, 1)]
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == []

    def test_shifted_boundary(self, base_sequence):
        """Test case: Entity with shifted boundary."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "O", "B-LOC"])

        evaluator = StrictEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 1
        assert result.incorrect == 1
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0)]
        assert result_indices.incorrect_indices == [(0, 1)]
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == []

    def test_extra_entity(self, base_sequence):
        """Test case: Extra entity in prediction."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "B-PER", "O", "B-LOC", "I-LOC"])

        evaluator = StrictEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 1
        assert result.incorrect == 1
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 1
        assert result_indices.correct_indices == [(0, 0)]
        assert result_indices.incorrect_indices == [(0, 2)]
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == [(0, 1)]


class TestEntityTypeEvaluation:
    """Test cases for entity type evaluation strategy."""

    def test_perfect_match(self, base_sequence):
        """Test case: Perfect match of all entities."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(base_sequence)

        evaluator = EntityTypeEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG"])

        assert result.correct == 2
        assert result.incorrect == 0
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0), (0, 1)]
        assert result_indices.incorrect_indices == []
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == []

    def test_missed_entity(self, base_sequence):
        """Test case: One entity is missed in prediction."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "O", "O"])

        evaluator = EntityTypeEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG"])

        assert result.correct == 1
        assert result.incorrect == 0
        assert result.partial == 0
        assert result.missed == 1
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0)]
        assert result_indices.incorrect_indices == []
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == [(0, 1)]
        assert result_indices.spurious_indices == []

    def test_wrong_label(self, base_sequence):
        """Test case: Entity with wrong label."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "B-LOC", "I-LOC"])

        evaluator = EntityTypeEvaluation()
        result, _ = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 1
        assert result.incorrect == 1
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 0

    def test_wrong_boundary(self, base_sequence):
        """Test case: Entity with wrong boundary."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "B-LOC", "O"])

        evaluator = EntityTypeEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 1
        assert result.incorrect == 1
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0)]
        assert result_indices.incorrect_indices == [(0, 1)]
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == []

    def test_shifted_boundary(self, base_sequence):
        """Test case: Entity with shifted boundary."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "O", "B-LOC"])

        evaluator = EntityTypeEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 1
        assert result.incorrect == 1
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0)]
        assert result_indices.incorrect_indices == [(0, 1)]
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == []

    def test_extra_entity(self, base_sequence):
        """Test case: Extra entity in prediction."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "B-PER", "O", "B-LOC", "I-LOC"])

        evaluator = EntityTypeEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 1
        assert result.incorrect == 1
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 1
        assert result_indices.correct_indices == [(0, 0)]
        assert result_indices.incorrect_indices == [(0, 2)]
        assert result_indices.spurious_indices == [(0, 1)]
        assert result_indices.missed_indices == []
        assert result_indices.partial_indices == []


class TestExactEvaluation:
    """Test cases for exact evaluation strategy."""

    def test_perfect_match(self, base_sequence):
        """Test case: Perfect match of all entities."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(base_sequence)

        evaluator = ExactEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG"])

        assert result.correct == 2
        assert result.incorrect == 0
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0), (0, 1)]
        assert result_indices.incorrect_indices == []
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == []

    def test_missed_entity(self, base_sequence):
        """Test case: One entity is missed in prediction."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "O", "O"])

        evaluator = ExactEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG"])

        assert result.correct == 1
        assert result.incorrect == 0
        assert result.partial == 0
        assert result.missed == 1
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0)]
        assert result_indices.incorrect_indices == []
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == [(0, 1)]
        assert result_indices.spurious_indices == []

    def test_wrong_label(self, base_sequence):
        """Test case: Entity with wrong label."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "B-LOC", "I-LOC"])

        evaluator = ExactEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 2
        assert result.incorrect == 0
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0), (0, 1)]
        assert result_indices.incorrect_indices == []
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == []

    def test_wrong_boundary(self, base_sequence):
        """Test case: Entity with wrong boundary."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "B-LOC", "O"])

        evaluator = ExactEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 1
        assert result.incorrect == 1
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0)]
        assert result_indices.incorrect_indices == [(0, 1)]
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == []

    def test_shifted_boundary(self, base_sequence):
        """Test case: Entity with shifted boundary."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "O", "B-LOC"])

        evaluator = ExactEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 1
        assert result.incorrect == 1
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0)]
        assert result_indices.incorrect_indices == [(0, 1)]
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == []

    def test_extra_entity(self, base_sequence):
        """Test case: Extra entity in prediction."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "B-PER", "O", "B-LOC", "I-LOC"])

        evaluator = ExactEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 2
        assert result.incorrect == 0
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 1
        assert result_indices.correct_indices == [(0, 0), (0, 2)]
        assert result_indices.incorrect_indices == []
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == [(0, 1)]


class TestPartialEvaluation:
    """Test cases for partial evaluation strategy."""

    def test_perfect_match(self, base_sequence):
        """Test case: Perfect match of all entities."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(base_sequence)

        evaluator = PartialEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG"])

        assert result.correct == 2
        assert result.incorrect == 0
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0), (0, 1)]
        assert result_indices.incorrect_indices == []
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == []

    def test_missed_entity(self, base_sequence):
        """Test case: One entity is missed in prediction."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "O", "O"])

        evaluator = PartialEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG"])

        assert result.correct == 1
        assert result.incorrect == 0
        assert result.partial == 0
        assert result.missed == 1
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0)]
        assert result_indices.incorrect_indices == []
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == [(0, 1)]
        assert result_indices.spurious_indices == []

    def test_wrong_label(self, base_sequence):
        """Test case: Entity with wrong label."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "B-LOC", "I-LOC"])

        evaluator = PartialEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 2
        assert result.incorrect == 0
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0), (0, 1)]
        assert result_indices.incorrect_indices == []
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == []

    def test_wrong_boundary(self, base_sequence):
        """Test case: Entity with wrong boundary."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "B-LOC", "O"])

        evaluator = PartialEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 1
        assert result.incorrect == 0
        assert result.partial == 1
        assert result.missed == 0
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0)]
        assert result_indices.incorrect_indices == []
        assert result_indices.partial_indices == [(0, 1)]
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == []

    def test_shifted_boundary(self, base_sequence):
        """Test case: Entity with shifted boundary."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "O", "O", "O", "B-LOC"])

        evaluator = PartialEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 1
        assert result.incorrect == 0
        assert result.partial == 1
        assert result.missed == 0
        assert result.spurious == 0
        assert result_indices.correct_indices == [(0, 0)]
        assert result_indices.incorrect_indices == []
        assert result_indices.partial_indices == [(0, 1)]
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == []

    def test_extra_entity(self, base_sequence):
        """Test case: Extra entity in prediction."""
        true = create_entities_from_bio(base_sequence)
        pred = create_entities_from_bio(["O", "B-PER", "I-PER", "O", "B-PER", "O", "B-LOC", "I-LOC"])

        evaluator = PartialEvaluation()
        result, result_indices = evaluator.evaluate(true, pred, ["PER", "ORG", "LOC"])

        assert result.correct == 2
        assert result.incorrect == 0
        assert result.partial == 0
        assert result.missed == 0
        assert result.spurious == 1
        assert result_indices.correct_indices == [(0, 0), (0, 2)]
        assert result_indices.incorrect_indices == []
        assert result_indices.partial_indices == []
        assert result_indices.missed_indices == []
        assert result_indices.spurious_indices == [(0, 1)]



import pytest

from nervaluate.loaders import ConllLoader, ListLoader, DictLoader


def test_conll_loader():
    """Test CoNLL format loader."""
    true_conll = (
        "word\tO\nword\tO\nword\tO\nword\tO\nword\tO\nword\tO\n\n"
        "word\tO\nword\tO\nword\tB-ORG\nword\tI-ORG\nword\tO\nword\tO\n\n"
        "word\tO\nword\tO\nword\tB-MISC\nword\tI-MISC\nword\tO\nword\tO\n\n"
        "word\tB-MISC\nword\tI-MISC\nword\tI-MISC\nword\tI-MISC\nword\tI-MISC\nword\tI-MISC\n"
    )

    pred_conll = (
        "word\tO\nword\tO\nword\tB-PER\nword\tI-PER\nword\tO\nword\tO\n\n"
        "word\tO\nword\tO\nword\tB-ORG\nword\tI-ORG\nword\tO\nword\tO\n\n"
        "word\tO\nword\tO\nword\tB-MISC\nword\tI-MISC\nword\tO\nword\tO\n\n"
        "word\tB-MISC\nword\tI-MISC\nword\tI-MISC\nword\tI-MISC\nword\tI-MISC\nword\tI-MISC\n"
    )

    loader = ConllLoader()
    true_entities = loader.load(true_conll)
    pred_entities = loader.load(pred_conll)

    # Test true entities
    assert len(true_entities) == 4  # Four documents
    assert len(true_entities[0]) == 0  # First document has no entities (all O tags)
    assert len(true_entities[1]) == 1  # Second document has 1 entity (ORG)
    assert len(true_entities[2]) == 1  # Third document has 1 entity (MISC)
    assert len(true_entities[3]) == 1  # Fourth document has 1 entity (MISC)

    # Check first entity in second document
    assert true_entities[1][0].label == "ORG"
    assert true_entities[1][0].start == 2
    assert true_entities[1][0].end == 3

    # Test pred entities
    assert len(pred_entities) == 4  # Four documents
    assert len(pred_entities[0]) == 1  # First document has 1 entity (PER)
    assert len(pred_entities[1]) == 1  # Second document has 1 entity (ORG)
    assert len(pred_entities[2]) == 1  # Third document has 1 entity (MISC)
    assert len(pred_entities[3]) == 1  # Fourth document has 1 entity (MISC)

    # Check first entity in first document
    assert pred_entities[0][0].label == "PER"
    assert pred_entities[0][0].start == 2
    assert pred_entities[0][0].end == 3

    # Test empty document handling
    empty_doc = "word\tO\nword\tO\nword\tO\n\n"
    empty_entities = loader.load(empty_doc)
    assert len(empty_entities) == 1  # One document
    assert len(empty_entities[0]) == 0  # Empty list for document with only O tags


def test_list_loader():
    """Test list format loader."""
    true_list = [
        ["O", "O", "O", "O", "O", "O"],
        ["O", "O", "B-ORG", "I-ORG", "O", "O"],
        ["O", "O", "B-MISC", "I-MISC", "O", "O"],
        ["B-MISC", "I-MISC", "I-MISC", "I-MISC", "I-MISC", "I-MISC"],
    ]

    pred_list = [
        ["O", "O", "B-PER", "I-PER", "O", "O"],
        ["O", "O", "B-ORG", "I-ORG", "O", "O"],
        ["O", "O", "B-MISC", "I-MISC", "O", "O"],
        ["B-MISC", "I-MISC", "I-MISC", "I-MISC", "I-MISC", "I-MISC"],
    ]

    loader = ListLoader()
    true_entities = loader.load(true_list)
    pred_entities = loader.load(pred_list)

    # Test true entities
    assert len(true_entities) == 4  # Four documents
    assert len(true_entities[0]) == 0  # First document has no entities (all O tags)
    assert len(true_entities[1]) == 1  # Second document has 1 entity (ORG)
    assert len(true_entities[2]) == 1  # Third document has 1 entity (MISC)
    assert len(true_entities[3]) == 1  # Fourth document has 1 entity (MISC)

    # Check no entities in the first document
    assert len(true_entities[0]) == 0

    # Check first entity in second document
    assert true_entities[1][0].label == "ORG"
    assert true_entities[1][0].start == 2
    assert true_entities[1][0].end == 3

    # Check only entity in the last document
    assert true_entities[3][0].label == "MISC"
    assert true_entities[3][0].start == 0
    assert true_entities[3][0].end == 5

    # Test pred entities
    assert len(pred_entities) == 4  # Four documents
    assert len(pred_entities[0]) == 1  # First document has 1 entity (PER)
    assert len(pred_entities[1]) == 1  # Second document has 1 entity (ORG)
    assert len(pred_entities[2]) == 1  # Third document has 1 entity (MISC)
    assert len(pred_entities[3]) == 1  # Fourth document has 1 entity (MISC)

    # Check first entity in first document
    assert pred_entities[0][0].label == "PER"
    assert pred_entities[0][0].start == 2
    assert pred_entities[0][0].end == 3

    # Test empty document handling
    empty_doc = [["O", "O", "O"]]
    empty_entities = loader.load(empty_doc)
    assert len(empty_entities) == 1  # One document
    assert len(empty_entities[0]) == 0  # Empty list for document with only O tags


def test_dict_loader():
    """Test dictionary format loader."""
    true_prod = [
        [],
        [{"label": "ORG", "start": 2, "end": 3}],
        [{"label": "MISC", "start": 2, "end": 3}],
        [{"label": "MISC", "start": 0, "end": 5}],
    ]

    pred_prod = [
        [{"label": "PER", "start": 2, "end": 3}],
        [{"label": "ORG", "start": 2, "end": 3}],
        [{"label": "MISC", "start": 2, "end": 3}],
        [{"label": "MISC", "start": 0, "end": 5}],
    ]

    loader = DictLoader()
    true_entities = loader.load(true_prod)
    pred_entities = loader.load(pred_prod)

    # Test true entities
    assert len(true_entities) == 4  # Four documents
    assert len(true_entities[0]) == 0  # First document has no entities
    assert len(true_entities[1]) == 1  # Second document has 1 entity (ORG)
    assert len(true_entities[2]) == 1  # Third document has 1 entity (MISC)
    assert len(true_entities[3]) == 1  # Fourth document has 1 entity (MISC)

    # Check first entity in second document
    assert true_entities[1][0].label == "ORG"
    assert true_entities[1][0].start == 2
    assert true_entities[1][0].end == 3

    # Check only entity in the last document
    assert true_entities[3][0].label == "MISC"
    assert true_entities[3][0].start == 0
    assert true_entities[3][0].end == 5

    # Test pred entities
    assert len(pred_entities) == 4  # Four documents
    assert len(pred_entities[0]) == 1  # First document has 1 entity (PER)
    assert len(pred_entities[1]) == 1  # Second document has 1 entity (ORG)
    assert len(pred_entities[2]) == 1  # Third document has 1 entity (MISC)
    assert len(pred_entities[3]) == 1  # Fourth document has 1 entity (MISC)

    # Check first entity in first document
    assert pred_entities[0][0].label == "PER"
    assert pred_entities[0][0].start == 2
    assert pred_entities[0][0].end == 3

    # Test empty document handling
    empty_doc = [[]]
    empty_entities = loader.load(empty_doc)
    assert len(empty_entities) == 1  # One document
    assert len(empty_entities[0]) == 0  # Empty list for empty document


def test_loader_with_empty_input():
    """Test loaders with empty input."""
    # Test ConllLoader with empty string
    conll_loader = ConllLoader()
    entities = conll_loader.load("")
    assert len(entities) == 0

    # Test ListLoader with empty list
    list_loader = ListLoader()
    entities = list_loader.load([])
    assert len(entities) == 0

    # Test DictLoader with empty list
    dict_loader = DictLoader()
    entities = dict_loader.load([])
    assert len(entities) == 0


def test_loader_with_invalid_data():
    """Test loaders with invalid data."""
    with pytest.raises(Exception):
        ConllLoader().load("invalid\tdata")

    with pytest.raises(Exception):
        ListLoader().load([["invalid"]])

    with pytest.raises(Exception):
        DictLoader().load([[{"invalid": "data"}]])



from dataclasses import dataclass
from typing import List, Tuple


@dataclass
class Entity:
    """Represents a named entity with its position and label."""

    label: str
    start: int
    end: int

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Entity):
            return NotImplemented
        return self.label == other.label and self.start == other.start and self.end == other.end

    def __hash__(self) -> int:
        return hash((self.label, self.start, self.end))


@dataclass
class EvaluationResult:
    """Represents the evaluation metrics for a single entity type or overall."""

    correct: int = 0
    incorrect: int = 0
    partial: int = 0
    missed: int = 0
    spurious: int = 0
    precision: float = 0.0
    recall: float = 0.0
    f1: float = 0.0
    actual: int = 0
    possible: int = 0

    def compute_metrics(self, partial_or_type: bool = False) -> None:
        """Compute precision, recall and F1 score."""
        self.actual = self.correct + self.incorrect + self.partial + self.spurious
        self.possible = self.correct + self.incorrect + self.partial + self.missed

        if partial_or_type:
            precision = (self.correct + 0.5 * self.partial) / self.actual if self.actual > 0 else 0
            recall = (self.correct + 0.5 * self.partial) / self.possible if self.possible > 0 else 0
        else:
            precision = self.correct / self.actual if self.actual > 0 else 0
            recall = self.correct / self.possible if self.possible > 0 else 0

        self.precision = precision
        self.recall = recall
        self.f1 = 2 * (precision * recall) / (precision + recall) if (precision + recall) > 0 else 0


@dataclass
class EvaluationIndices:
    """Represents the indices of entities in different evaluation categories."""

    correct_indices: List[Tuple[int, int]] = None  # type: ignore
    incorrect_indices: List[Tuple[int, int]] = None  # type: ignore
    partial_indices: List[Tuple[int, int]] = None  # type: ignore
    missed_indices: List[Tuple[int, int]] = None  # type: ignore
    spurious_indices: List[Tuple[int, int]] = None  # type: ignore

    def __post_init__(self) -> None:
        if self.correct_indices is None:
            self.correct_indices = []
        if self.incorrect_indices is None:
            self.incorrect_indices = []
        if self.partial_indices is None:
            self.partial_indices = []
        if self.missed_indices is None:
            self.missed_indices = []
        if self.spurious_indices is None:
            self.spurious_indices = []



def split_list(token: list[str], split_chars: list[str] | None = None) -> list[list[str]]:
    """
    Split a list into sublists based on a list of split characters.

    If split_chars is None, the list is split on empty strings.

    :param token: The list to split.
    :param split_chars: The characters to split on.

    :returns:
        A list of lists.
    """
    if split_chars is None:
        split_chars = [""]
    out = []
    chunk = []
    for i, item in enumerate(token):
        if item not in split_chars:
            chunk.append(item)
            if i + 1 == len(token):
                out.append(chunk)
        else:
            out.append(chunk)
            chunk = []
    return out


def conll_to_spans(doc: str) -> list[list[dict]]:
    """
    Convert a CoNLL-formatted string to a list of spans.

    :param doc: The CoNLL-formatted string.

    :returns:
        A list of spans.
    """
    out = []
    doc_parts = split_list(doc.split("\n"), split_chars=None)

    for example in doc_parts:
        labels = []
        for token in example:
            token_parts = token.split("\t")
            label = token_parts[1]
            labels.append(label)
        out.append(labels)

    spans = list_to_spans(out)

    return spans


def list_to_spans(doc: list[list[str]]) -> list[list[dict]]:
    """
    Convert a list of tags to a list of spans.

    :param doc: The list of tags.

    :returns:
        A list of spans.
    """
    spans = [collect_named_entities(tokens) for tokens in doc]
    return spans


def collect_named_entities(tokens: list[str]) -> list[dict]:
    """
    Creates a list of Entity named-tuples, storing the entity type and the start and end offsets of the entity.

    :param tokens: a list of tags

    :returns:
        A list of Entity named-tuples.
    """

    named_entities = []
    start_offset = None
    end_offset = None
    ent_type = None

    for offset, token_tag in enumerate(tokens):
        if token_tag == "O":
            if ent_type is not None and start_offset is not None:
                end_offset = offset - 1
                named_entities.append({"label": ent_type, "start": start_offset, "end": end_offset})
                start_offset = None
                end_offset = None
                ent_type = None

        elif ent_type is None:
            ent_type = token_tag[2:]
            start_offset = offset

        elif ent_type != token_tag[2:] or (ent_type == token_tag[2:] and token_tag[:1] == "B"):
            end_offset = offset - 1
            named_entities.append({"label": ent_type, "start": start_offset, "end": end_offset})

            # start of a new entity
            ent_type = token_tag[2:]
            start_offset = offset
            end_offset = None

    # Catches an entity that goes up until the last token
    if ent_type is not None and start_offset is not None and end_offset is None:
        named_entities.append({"label": ent_type, "start": start_offset, "end": len(tokens) - 1})

    return named_entities


def find_overlap(true_range: range, pred_range: range) -> set:
    """
    Find the overlap between two ranges.

    :param true_range: The true range.
    :param pred_range: The predicted range.

    :returns:
        A set of overlapping values.

    Examples:
        >>> find_overlap(range(1, 3), range(2, 4))
        {2}
        >>> find_overlap(range(1, 3), range(3, 5))
        set()
    """

    true_set = set(true_range)
    pred_set = set(pred_range)
    overlaps = true_set.intersection(pred_set)

    return overlaps


def clean_entities(ent: dict) -> dict:
    """
    Returns just the useful keys if additional keys are present in the entity
    dict.

    This may happen if passing a list of spans directly from prodigy, which
    typically may include 'token_start' and 'token_end'.
    """
    return {"start": ent["start"], "end": ent["end"], "label": ent["label"]}



from abc import ABC, abstractmethod
from typing import List, Dict, Any

from .entities import Entity


class DataLoader(ABC):
    """Abstract base class for data loaders."""

    @abstractmethod
    def load(self, data: Any) -> List[List[Entity]]:
        """Load data into a list of entity lists."""


class ConllLoader(DataLoader):
    """Loader for CoNLL format data."""

    def load(self, data: str) -> List[List[Entity]]:  # pylint: disable=too-many-branches
        """Load CoNLL format data into a list of Entity lists."""
        if not isinstance(data, str):
            raise ValueError("ConllLoader expects string input")

        if not data:
            return []

        result: List[List[Entity]] = []
        # Strip trailing whitespace and newlines to avoid empty documents
        documents = data.rstrip().split("\n\n")

        for doc in documents:
            if not doc.strip():
                result.append([])
                continue

            current_doc = []
            start_offset = None
            end_offset = None
            ent_type = None
            has_entities = False

            for offset, line in enumerate(doc.split("\n")):
                if not line.strip():
                    continue

                parts = line.split("\t")
                if len(parts) < 2:
                    raise ValueError(f"Invalid CoNLL format: line '{line}' does not contain a tab separator")

                token_tag = parts[1]

                if token_tag == "O":
                    if ent_type is not None and start_offset is not None:
                        end_offset = offset - 1
                        if isinstance(start_offset, int) and isinstance(end_offset, int):
                            current_doc.append(Entity(label=ent_type, start=start_offset, end=end_offset))
                        start_offset = None
                        end_offset = None
                        ent_type = None

                elif ent_type is None:
                    if not (token_tag.startswith("B-") or token_tag.startswith("I-")):
                        raise ValueError(f"Invalid tag format: {token_tag}")
                    ent_type = token_tag[2:]  # Remove B- or I- prefix
                    start_offset = offset
                    has_entities = True

                elif ent_type != token_tag[2:] or (ent_type == token_tag[2:] and token_tag[:1] == "B"):
                    end_offset = offset - 1
                    if isinstance(start_offset, int) and isinstance(end_offset, int):
                        current_doc.append(Entity(label=ent_type, start=start_offset, end=end_offset))

                    # start of a new entity
                    if not (token_tag.startswith("B-") or token_tag.startswith("I-")):
                        raise ValueError(f"Invalid tag format: {token_tag}")
                    ent_type = token_tag[2:]
                    start_offset = offset
                    end_offset = None
                    has_entities = True

            # Catches an entity that goes up until the last token
            if ent_type is not None and start_offset is not None and end_offset is None:
                if isinstance(start_offset, int):
                    current_doc.append(Entity(label=ent_type, start=start_offset, end=len(doc.split("\n")) - 1))
                has_entities = True

            result.append(current_doc if has_entities else [])

        return result


class ListLoader(DataLoader):
    """Loader for list format data."""

    def load(self, data: List[List[str]]) -> List[List[Entity]]:  # pylint: disable=too-many-branches
        """Load list format data into a list of entity lists."""
        if not isinstance(data, list):
            raise ValueError("ListLoader expects list input")

        if not data:
            return []

        result = []

        for doc in data:
            if not isinstance(doc, list):
                raise ValueError("Each document must be a list of tags")

            current_doc = []
            start_offset = None
            end_offset = None
            ent_type = None

            for offset, token_tag in enumerate(doc):
                if not isinstance(token_tag, str):
                    raise ValueError(f"Invalid tag type: {type(token_tag)}")

                if token_tag == "O":
                    if ent_type is not None and start_offset is not None:
                        end_offset = offset - 1
                        if isinstance(start_offset, int) and isinstance(end_offset, int):
                            current_doc.append(Entity(label=ent_type, start=start_offset, end=end_offset))
                        start_offset = None
                        end_offset = None
                        ent_type = None

                elif ent_type is None:
                    if not (token_tag.startswith("B-") or token_tag.startswith("I-")):
                        raise ValueError(f"Invalid tag format: {token_tag}")
                    ent_type = token_tag[2:]  # Remove B- or I- prefix
                    start_offset = offset

                elif ent_type != token_tag[2:] or (ent_type == token_tag[2:] and token_tag[:1] == "B"):
                    end_offset = offset - 1
                    if isinstance(start_offset, int) and isinstance(end_offset, int):
                        current_doc.append(Entity(label=ent_type, start=start_offset, end=end_offset))

                    # start of a new entity
                    if not (token_tag.startswith("B-") or token_tag.startswith("I-")):
                        raise ValueError(f"Invalid tag format: {token_tag}")
                    ent_type = token_tag[2:]
                    start_offset = offset
                    end_offset = None

            # Catches an entity that goes up until the last token
            if ent_type is not None and start_offset is not None and end_offset is None:
                if isinstance(start_offset, int):
                    current_doc.append(Entity(label=ent_type, start=start_offset, end=len(doc) - 1))

            result.append(current_doc)

        return result


class DictLoader(DataLoader):
    """Loader for dictionary format data."""

    def load(self, data: List[List[Dict[str, Any]]]) -> List[List[Entity]]:
        """Load dictionary format data into a list of entity lists."""
        if not isinstance(data, list):
            raise ValueError("DictLoader expects list input")

        if not data:
            return []

        result = []

        for doc in data:
            if not isinstance(doc, list):
                raise ValueError("Each document must be a list of entity dictionaries")

            current_doc = []
            for entity in doc:
                if not isinstance(entity, dict):
                    raise ValueError(f"Invalid entity type: {type(entity)}")

                required_keys = {"label", "start", "end"}
                if not all(key in entity for key in required_keys):
                    raise ValueError(f"Entity missing required keys: {required_keys}")

                if not isinstance(entity["label"], str):
                    raise ValueError("Entity label must be a string")

                if not isinstance(entity["start"], int) or not isinstance(entity["end"], int):
                    raise ValueError("Entity start and end must be integers")

                current_doc.append(Entity(label=entity["label"], start=entity["start"], end=entity["end"]))
            result.append(current_doc)

        return result
